/* 
 * Copyright 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995, 1996,
 * 1997, 2001, 2004. The Regents of the University of California 
 * (Regents). All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research, and not-for-profit purposes,
 * without fee and without a signed licensing agreement, is hereby
 * granted, provided that the above copyright notice, this paragraph and
 * the following three paragraphs appear in all copies, modifications, and
 * distributions. Contact The Office of Technology Licensing, UC
 * Berkeley, 2150 Shattuck Avenue, Suite 510, Berkeley, CA 94720-1620,
 * (510) 643-7201, for commercial licensing opportunities. 
 *
 * This code is derived from the Stanford/CMU enet packet filter,
 * distributed as part of 4.3BSD, and code contributed to Berkeley by
 * Steven McCanne and Van Jacobson both of Lawrence Berkeley Laboratory,
 * and Andrew Begel of the Berkeley Computer Science Division of the
 * Electrical Engineering and Computer Science Department.
 * 
 * IN NO EVENT SHALL REGENTS BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
 * REGENTS HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * REGENTS SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE. THE SOFTWARE AND ACCOMPANYING DOCUMENTATION, IF
 * ANY, PROVIDED HEREUNDER IS PROVIDED "AS IS".  REGENTS HAS NO
 * OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
 * MODIFICATIONS.
 */
#ifndef lint
static char rcsid[] =
    "@(#) $Header: /usr/src/mash/repository/bpfl/gencode.c,v 1.6 1997/11/20 17:39:27 mccanne Exp $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>

#include <stdlib.h>
#include <memory.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdarg.h>

#include "bpf-vm.h"
#include "gencode.h"

/*XXX*/
#include "pcap-namedb.h"

#ifndef ETHERTYPE_REVARP
#define ETHERTYPE_REVARP	0x8035
#endif
#ifndef	ETHERTYPE_MOPDL
#define	ETHERTYPE_MOPDL		0x6001
#endif
#ifndef	ETHERTYPE_MOPRC
#define	ETHERTYPE_MOPRC		0x6002
#endif
#ifndef ETHERTYPE_NS
#define ETHERTYPE_NS		0x0600
#endif
#ifndef	ETHERTYPE_SPRITE
#define	ETHERTYPE_SPRITE	0x0500
#endif
#ifndef ETHERTYPE_TRAIL
#define ETHERTYPE_TRAIL		0x1000
#endif
#ifndef	ETHERTYPE_DN
#define	ETHERTYPE_DN		0x6003
#endif
#ifndef	ETHERTYPE_LAT
#define	ETHERTYPE_LAT		0x6004
#endif
#ifndef ETHERTYPE_SCA
#define ETHERTYPE_SCA		0x6007
#endif
#ifndef ETHERTYPE_REVARP
#define ETHERTYPE_REVARP	0x8035
#endif
#ifndef	ETHERTYPE_LANBRIDGE
#define	ETHERTYPE_LANBRIDGE	0x8038
#endif
#ifndef	ETHERTYPE_DECDNS
#define	ETHERTYPE_DECDNS	0x803c
#endif
#ifndef	ETHERTYPE_DECDTS
#define	ETHERTYPE_DECDTS	0x803e
#endif
#ifndef	ETHERTYPE_VEXP
#define	ETHERTYPE_VEXP		0x805b
#endif
#ifndef	ETHERTYPE_VPROD
#define	ETHERTYPE_VPROD		0x805c
#endif
#ifndef ETHERTYPE_ATALK
#define ETHERTYPE_ATALK		0x809b
#endif
#ifndef ETHERTYPE_AARP
#define ETHERTYPE_AARP		0x80f3
#endif
#ifndef	ETHERTYPE_LOOPBACK
#define	ETHERTYPE_LOOPBACK	0x9000
#endif


#define JMP(c) ((c)|BPF_JMP|BPF_IMM)

static jmp_buf top_ctx;

/*XXX should pass control to method */
static char errorBuffer[256];
char* bpf_geterror() { return (errorBuffer); }

void
bpf_error(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	(void)vsprintf(errorBuffer, fmt, ap);
	va_end(ap);
	longjmp(top_ctx, 1);
	/* NOTREACHED */
}

static void init_linkinfo(void);

static int alloc_mem(void);
static int alloc_reg(void);
static void free_reg(int);

static struct block *root;

/*
 * We divy out chunks of memory rather than call malloc each time so
 * we don't have to worry about leaking memory.  It's probably
 * not a big deal if all this memory was wasted but it this ever
 * goes into a library that would probably not be a good idea.
 */
#define NCHUNKS 16
#define CHUNK0SIZE 1024
struct chunk {
	u_int n_left;
	void *m;
};

static struct chunk chunks[NCHUNKS];
static int cur_chunk;

static void *
newchunk(u_int n)
{
	struct chunk *cp;
	int k, size;

	/* XXX Round up to nearest long long. */
	n = (n + sizeof(long long) - 1) & ~(sizeof(long long) - 1);

	cp = &chunks[cur_chunk];
	if (n > cp->n_left) {
		++cp, k = ++cur_chunk;
		if (k >= NCHUNKS)
			bpf_error("out of memory");
		size = CHUNK0SIZE << k;
		cp->m = (void *)malloc(size);
		memset((char *)cp->m, 0, size);
		cp->n_left = size;
		if (n > size)
			bpf_error("out of memory");
	}
	cp->n_left -= n;
	return (void *)((char *)cp->m + cp->n_left);
}

static void
freechunks(void)
{
	int i;

	cur_chunk = 0;
	for (i = 0; i < NCHUNKS; ++i)
		if (chunks[i].m != NULL) {
			free(chunks[i].m);
			chunks[i].m = NULL;
		}
}

/*
 * A strdup whose allocations are freed after code generation is over.
 */
char *
sdup(s)
	char *s;
{
	int n = strlen(s) + 1;
	char *cp = newchunk(n);
	strcpy(cp, s);
	return (cp);
}

static inline struct block *
new_block(int code, int rd, int imm)
{
	struct block *p;

	p = (struct block *)newchunk(sizeof(*p));
	p->s.code = code;
	p->s.rd = rd;
	p->s.imm = imm;
	p->head = p;

	return p;
}

/*XXX*/
static inline struct slist *
new_stmt(int code, int ra, int rb, int rd, int imm)
{
	struct slist *p;

	p = (struct slist *)newchunk(sizeof(*p));
	p->s.code = code;
	p->s.ra = ra;
	p->s.rb = rb;
	p->s.rd = rd;
	p->s.imm = imm;

	return p;
}

static inline struct block *
gen_retblk(int v)
{
	return (new_block(BPF_RET|BPF_IMM, 0, v));
}

static inline void
syntax(void)
{
	bpf_error("syntax error in filter expression");
}

static u_int32_t netmask;
static int snaplen;

struct bpf_insn *
bpfl_compile(int* programLen,
	     char *buf, int optimize, u_int32_t mask,
	     int retval)
{
	struct bpf_insns* program;

	extern int n_errors;
	int len;

	n_errors = 0;
	root = NULL;
	if (setjmp(top_ctx)) {
		freechunks();
		return (0);
	}

	netmask = mask;
	snaplen = retval;

	lex_init(buf ? buf : "");

	/*XXX*/
	init_linkinfo();

	(void)bpfl_parse();

	if (n_errors)
		syntax();

	if (root == NULL)
		root = gen_retblk(snaplen);

	if (optimize) {
	  bpf_optimize(&root);
	  if (root == NULL) printf("root is null\n");
	  if (root == NULL ||
	      (root->s.code == (BPF_RET|BPF_IMM) && root->s.imm == 0))
	    bpf_error("expression rejects all packets");
	}
	else bpf_dont_optimize(&root);
	
	program = icode_to_fcode(root, &len);
	*programLen = len;

	freechunks();

	return (program);
}

/*
 * Backpatch the blocks in 'list' to 'target'.  The 'sense' field indicates
 * which of the jt and jf fields has been resolved and which is a pointer
 * back to another unresolved block (or nil).  At least one of the fields
 * in each block is already resolved.
 */
static void
backpatch(struct block *list, struct block *target)
{
	struct block *next;

	while (list) {
		if (!list->sense) {
			next = JT(list);
			JT(list) = target;
		} else {
			next = JF(list);
			JF(list) = target;
		}
		list = next;
	}
}

/*
 * Merge the lists in b0 and b1, using the 'sense' field to indicate
 * which of jt and jf is the link.
 */
static void
merge(struct block *b0, struct block *b1)
{
	register struct block **p = &b0;

	/* Find end of list. */
	while (*p)
		p = !((*p)->sense) ? &JT(*p) : &JF(*p);

	/* Concatenate the lists. */
	*p = b1;
}

void
finish_parse(p)
	struct block *p;
{
	backpatch(p, gen_retblk(snaplen));
	p->sense = !p->sense;
	backpatch(p, gen_retblk(0));
	root = p->head;
}

void
gen_and(b0, b1)
	struct block *b0, *b1;
{
	backpatch(b0, b1->head);
	b0->sense = !b0->sense;
	b1->sense = !b1->sense;
	merge(b1, b0);
	b1->sense = !b1->sense;
	b1->head = b0->head;
}

void
gen_or(b0, b1)
	struct block *b0, *b1;
{
	b0->sense = !b0->sense;
	backpatch(b0, b1->head);
	b0->sense = !b0->sense;
	merge(b1, b0);
	b1->head = b0->head;
}

void
gen_not(b)
	struct block *b;
{
	b->sense = !b->sense;
}

static struct block *
gen_cmp(u_int offset, u_int size, long v)
{
	struct slist *s;
	struct block *b;
	int reg;

	reg = alloc_reg();
	s = new_stmt(BPF_LD|BPF_ABS|size, 0, 0, reg, offset);
	b = new_block(JMP(BPF_JEQ), reg, v);
	b->stmts = s;

	return b;
}

static struct block *
gen_mcmp(u_int offset, u_int size, long v, u_long mask)
{
	struct block *b = gen_cmp(offset, size, v);
	struct slist *s;
	int reg, reg1;

	reg = b->stmts->s.rd;
	reg1 = alloc_reg();
	if (mask != 0xffffffff) {
		s = new_stmt(BPF_ALU|BPF_AND|BPF_IMM, reg, 0, reg1, mask);
		b->stmts->next = s;
		b->s.rd = reg1; // change the jump to use the right reg.
	}
	return b;
}

static struct block *
gen_bcmp(u_int offset, u_int size, u_char *v)
{
	struct block *b, *tmp;

	b = NULL;
	while (size >= 4) {
		u_char *p = &v[size - 4];
		long w = (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
		tmp = gen_cmp(offset + size - 4, BPF_W, w);
		if (b != NULL)
			gen_and(b, tmp);
		b = tmp;
		size -= 4;
	}
	while (size >= 2) {
		u_char *p = &v[size - 2];
		long w = (p[0] << 8) | p[1];
		tmp = gen_cmp(offset + size - 2, BPF_H, w);
		if (b != NULL)
			gen_and(b, tmp);
		b = tmp;
		size -= 2;
	}
	if (size > 0) {
		tmp = gen_cmp(offset, BPF_B, (long)v[0]);
		if (b != NULL)
			gen_and(b, tmp);
		b = tmp;
	}
	return b;
}

/*
 * Various configuration parameters for the link layer header
 * so that we can generate code for checking link layer fields
 * and skipping over the link layer header.  If any of the
 * parameters are irrelevant (e.g., type on a PPP link),
 * then the value is set to -1.
 */
static struct {
	int hdrlen;		/* header size (must be fixed size) */
	int addrlen;		/* size of a link address */
	int type_width;		/* width of type field (BPF_W etc) */
	int off_type;		/* byte offset to type field */
	int off_src;		/* byte offset to source address */
	int off_dst;		/* byte offset to dest address */
} linkinfo;


static void
init_linkinfo()
{
	printf("fix init_linkinfo()\n");

	/*XXX assume ethernet */

	linkinfo.hdrlen = 14;
	linkinfo.addrlen = 6;
	linkinfo.type_width = BPF_H;
	linkinfo.off_type = 12;
	linkinfo.off_src = 6;
	linkinfo.off_dst = 0;

#ifdef notdef
	switch (type) {

	case DLT_EN10MB:
		off_linktype = 12;
		off_nl = 14;
		return;

	case DLT_SLIP:
		/*
		 * SLIP doesn't have a link level type.  The 16 byte
		 * header is hacked into our SLIP driver.
		 */
		off_linktype = -1;
		off_nl = 16;
		return;

	case DLT_SLIP_BSDOS:
		/* XXX this may be the same as the DLT_PPP_BSDOS case */
		off_linktype = -1;
		/* XXX end */
		off_nl = 24;
		return;

	case DLT_NULL:
		off_linktype = -1;
		off_nl = 0;
		return;

	case DLT_PPP:
		off_linktype = 2;
		off_nl = 4;
		return;

	case DLT_PPP_BSDOS:
		off_linktype = 5;
		off_nl = 24;
		return;

	case DLT_FDDI:
		/*
		 * FDDI doesn't really have a link-level type field.
		 * We assume that SSAP = SNAP is being used and pick
		 * out the encapsulated Ethernet type.
		 */
		off_linktype = 19 + fddipad;
		off_nl = 21 + fddipad;
		return;

	case DLT_IEEE802:
		off_linktype = 20;
		off_nl = 22;
		return;

	case DLT_ATM_RFC1483:
		/*
		 * assume routed, non-ISO PDUs
		 * (i.e., LLC = 0xAA-AA-03, OUT = 0x00-00-00)
		 */
		off_linktype = 6;
		off_nl = 8;
		return;

	case DLT_RAW:
		off_linktype = -1;
		off_nl = 0;
		return;
	}
	bpf_error("unknown data link type 0x%x", linktype);
#endif
}

static struct block *
gen_uncond(int rsense)
{
	struct block *b;
	struct slist *s;
	int reg;

	reg = alloc_reg();
	s = new_stmt(BPF_LD|BPF_IMM, 0, 0, reg, !rsense);
	b = new_block(JMP(BPF_JEQ), reg, 0);
	b->stmts = s;

	return b;
}

static inline struct block *
gen_true(void)
{
	return gen_uncond(1);
}

static inline struct block *
gen_false(void)
{
	return gen_uncond(0);
}

static struct block *
gen_linktype(int proto)
{
	if (linkinfo.hdrlen < 0)
		return gen_true();

	/*XXX*/
#ifdef notdef
	case DLT_SLIP:
		return gen_false();

	case DLT_PPP:
		if (proto == ETHERTYPE_IP)
			proto = 0x0021;		/* XXX - need ppp.h defs */
		break;
	case DLT_PPP_BSDOS:
		switch (proto) {

		case ETHERTYPE_IP:
			b0 = gen_cmp(off_linktype, BPF_H, PPP_IP);
			b1 = gen_cmp(off_linktype, BPF_H, PPP_VJC);
			gen_or(b0, b1);
			b0 = gen_cmp(off_linktype, BPF_H, PPP_VJNC);
			gen_or(b1, b0);
			return b0;

		case ETHERTYPE_DN:
			proto = PPP_DECNET;
			break;

		case ETHERTYPE_ATALK:
			proto = PPP_APPLE;
			break;

		case ETHERTYPE_NS:
			proto = PPP_NS;
			break;
		}
		break;

	case DLT_NULL:
		/* XXX */
		if (proto == ETHERTYPE_IP)
			return (gen_cmp(0, BPF_W, (long)AF_INET));
		else
			return gen_false();
#endif
	return gen_cmp(linkinfo.off_type, linkinfo.type_width, (long)proto);
}

static struct block *
gen_hostop(u_long addr, u_long mask, int dir, int proto,
	   u_int src_off, u_int dst_off)
{
	struct block *b0, *b1;
	u_int offset;

	switch (dir) {

	case Q_SRC:
		offset = src_off;
		break;

	case Q_DST:
		offset = dst_off;
		break;

	case Q_AND:
		b0 = gen_hostop(addr, mask, Q_SRC, proto, src_off, dst_off);
		b1 = gen_hostop(addr, mask, Q_DST, proto, src_off, dst_off);
		gen_and(b0, b1);
		return b1;

	case Q_OR:
	case Q_DEFAULT:
		b0 = gen_hostop(addr, mask, Q_SRC, proto, src_off, dst_off);
		b1 = gen_hostop(addr, mask, Q_DST, proto, src_off, dst_off);
		gen_or(b0, b1);
		return b1;

	default:
		abort();
	}
	b0 = gen_linktype(proto);
	b1 = gen_mcmp(offset, BPF_W, (long)addr, mask);
	gen_and(b0, b1);
	return b1;
}

static struct block *
gen_ehostop(u_char *eaddr, int dir)
{
	struct block *b0, *b1;

	switch (dir) {
	case Q_SRC:
		return gen_bcmp(linkinfo.off_src, linkinfo.addrlen, eaddr);

	case Q_DST:
		return gen_bcmp(linkinfo.off_dst, linkinfo.addrlen, eaddr);

	case Q_AND:
		b0 = gen_ehostop(eaddr, Q_SRC);
		b1 = gen_ehostop(eaddr, Q_DST);
		gen_and(b0, b1);
		return b1;

	case Q_DEFAULT:
	case Q_OR:
		b0 = gen_ehostop(eaddr, Q_SRC);
		b1 = gen_ehostop(eaddr, Q_DST);
		gen_or(b0, b1);
		return b1;
	}
	abort();
	/* NOTREACHED */
}

/*
 * This is quite tricky because there may be pad bytes in front of the
 * DECNET header, and then there are two possible data packet formats that
 * carry both src and dst addresses, plus 5 packet types in a format that
 * carries only the src node, plus 2 types that use a different format and
 * also carry just the src node.
 *
 * Yuck.
 *
 * Instead of doing those all right, we just look for data packets with
 * 0 or 1 bytes of padding.  If you want to look at other packets, that
 * will require a lot more hacking.
 *
 * To add support for filtering on DECNET "areas" (network numbers)
 * one would want to add a "mask" argument to this routine.  That would
 * make the filter even more inefficient, although one could be clever
 * and not generate masking instructions if the mask is 0xFFFF.
 */
static struct block *
gen_dnhostop(u_long addr, int dir, u_int base_off)
{
	struct block *b0, *b1, *b2, *tmp;
	u_int offset_lh;	/* offset if long header is received */
	u_int offset_sh;	/* offset if short header is received */

	switch (dir) {

	case Q_DST:
		offset_sh = 1;	/* follows flags */
		offset_lh = 7;	/* flgs,darea,dsubarea,HIORD */
		break;

	case Q_SRC:
		offset_sh = 3;	/* follows flags, dstnode */
		offset_lh = 15;	/* flgs,darea,dsubarea,did,sarea,ssub,HIORD */
		break;

	case Q_AND:
		/* Inefficient because we do our Calvinball dance twice */
		b0 = gen_dnhostop(addr, Q_SRC, base_off);
		b1 = gen_dnhostop(addr, Q_DST, base_off);
		gen_and(b0, b1);
		return b1;

	case Q_OR:
	case Q_DEFAULT:
		/* Inefficient because we do our Calvinball dance twice */
		b0 = gen_dnhostop(addr, Q_SRC, base_off);
		b1 = gen_dnhostop(addr, Q_DST, base_off);
		gen_or(b0, b1);
		return b1;

	default:
		abort();
	}
	b0 = gen_linktype(ETHERTYPE_DN);
	/* Check for pad = 1, long header case */
	tmp = gen_mcmp(base_off + 2, BPF_H,
				(long)ntohs(0x0681), (long)ntohs(0x07FF));
	b1 = gen_cmp(base_off + 2 + 1 + offset_lh, BPF_H, (long)ntohs(addr));
	gen_and(tmp, b1);
	/* Check for pad = 0, long header case */
	tmp = gen_mcmp(base_off + 2, BPF_B, (long)0x06, (long)0x7);
	b2 = gen_cmp(base_off + 2 + offset_lh, BPF_H, (long)ntohs(addr));
	gen_and(tmp, b2);
	gen_or(b2, b1);
	/* Check for pad = 1, short header case */
	tmp = gen_mcmp(base_off + 2, BPF_H,
				(long)ntohs(0x0281), (long)ntohs(0x07FF));
	b2 = gen_cmp(base_off + 2 + 1 + offset_sh, BPF_H, (long)ntohs(addr));
	gen_and(tmp, b2);
	gen_or(b2, b1);
	/* Check for pad = 0, short header case */
	tmp = gen_mcmp(base_off + 2, BPF_B, (long)0x02, (long)0x7);
	b2 = gen_cmp(base_off + 2 + offset_sh, BPF_H, (long)ntohs(addr));
	gen_and(tmp, b2);
	gen_or(b2, b1);

	/* Combine with test for linktype */
	gen_and(b0, b1);
	return b1;
}

static struct block *
gen_host(u_long addr, u_long mask, int proto, int dir)
{
	struct block *b0, *b1;

	switch (proto) {

	case Q_DEFAULT:
		b0 = gen_host(addr, mask, Q_IP, dir);
		b1 = gen_host(addr, mask, Q_ARP, dir);
		gen_or(b0, b1);
		b0 = gen_host(addr, mask, Q_RARP, dir);
		gen_or(b1, b0);
		return b0;

	case Q_IP:
		return gen_hostop(addr, mask, dir, ETHERTYPE_IP,
				  linkinfo.hdrlen + 12,
				  linkinfo.hdrlen + 16);

	case Q_RARP:
		return gen_hostop(addr, mask, dir, ETHERTYPE_REVARP,
				  linkinfo.hdrlen + 14, linkinfo.hdrlen + 24);

	case Q_ARP:
		return gen_hostop(addr, mask, dir, ETHERTYPE_ARP,
				  linkinfo.hdrlen + 14, linkinfo.hdrlen + 24);

	case Q_TCP:
		bpf_error("'tcp' modifier applied to host");

	case Q_UDP:
		bpf_error("'udp' modifier applied to host");

	case Q_ICMP:
		bpf_error("'icmp' modifier applied to host");

	case Q_IGMP:
		bpf_error("'igmp' modifier applied to host");

	case Q_IGRP:
		bpf_error("'igrp' modifier applied to host");

	case Q_ATALK:
		bpf_error("ATALK host filtering not implemented");

	case Q_DECNET:
		return gen_dnhostop(addr, dir, linkinfo.hdrlen);

	case Q_SCA:
		bpf_error("SCA host filtering not implemented");

	case Q_LAT:
		bpf_error("LAT host filtering not implemented");

	case Q_MOPDL:
		bpf_error("MOPDL host filtering not implemented");

	case Q_MOPRC:
		bpf_error("MOPRC host filtering not implemented");

	default:
		abort();
	}
	/* NOTREACHED */
}

static struct block *
gen_gateway(u_char *eaddr, u_long **alist, int proto, int dir)
{
	struct block *b0, *b1, *tmp;

	if (dir != 0)
		bpf_error("direction applied to 'gateway'");

	switch (proto) {
	case Q_DEFAULT:
	case Q_IP:
	case Q_ARP:
	case Q_RARP:
		b0 = gen_ehostop(eaddr, Q_OR);
		b1 = gen_host(**alist++, 0xffffffffL, proto, Q_OR);
		while (*alist) {
			tmp = gen_host(**alist++, 0xffffffffL, proto, Q_OR);
			gen_or(b1, tmp);
			b1 = tmp;
		}
		gen_not(b1);
		gen_and(b0, b1);
		return b1;
	}
	bpf_error("illegal modifier of 'gateway'");
	/* NOTREACHED */
}

struct block *
gen_proto_abbrev(int proto)
{
	struct block *b0, *b1;

	switch (proto) {

	case Q_TCP:
		b0 = gen_linktype(ETHERTYPE_IP);
		b1 = gen_cmp(linkinfo.hdrlen + 9, BPF_B, (long)IPPROTO_TCP);
		gen_and(b0, b1);
		break;

	case Q_UDP:
		b0 =  gen_linktype(ETHERTYPE_IP);
		b1 = gen_cmp(linkinfo.hdrlen + 9, BPF_B, (long)IPPROTO_UDP);
		gen_and(b0, b1);
		break;

	case Q_ICMP:
		b0 =  gen_linktype(ETHERTYPE_IP);
		b1 = gen_cmp(linkinfo.hdrlen + 9, BPF_B, (long)IPPROTO_ICMP);
		gen_and(b0, b1);
		break;

	case Q_IGMP:
		b0 =  gen_linktype(ETHERTYPE_IP);
		b1 = gen_cmp(linkinfo.hdrlen + 9, BPF_B, (long)2);
		gen_and(b0, b1);
		break;

#ifndef	IPPROTO_IGRP
#define	IPPROTO_IGRP	9
#endif
	case Q_IGRP:
		b0 =  gen_linktype(ETHERTYPE_IP);
		b1 = gen_cmp(linkinfo.hdrlen + 9, BPF_B, (long)IPPROTO_IGRP);
		gen_and(b0, b1);
		break;

	case Q_IP:
		b1 =  gen_linktype(ETHERTYPE_IP);
		break;

	case Q_ARP:
		b1 =  gen_linktype(ETHERTYPE_ARP);
		break;

	case Q_RARP:
		b1 =  gen_linktype(ETHERTYPE_REVARP);
		break;

	case Q_LINK:
		bpf_error("link layer applied in wrong context");

	case Q_ATALK:
		b1 =  gen_linktype(ETHERTYPE_ATALK);
		break;

	case Q_DECNET:
		b1 =  gen_linktype(ETHERTYPE_DN);
		break;

	case Q_SCA:
		b1 =  gen_linktype(ETHERTYPE_SCA);
		break;

	case Q_LAT:
		b1 =  gen_linktype(ETHERTYPE_LAT);
		break;

	case Q_MOPDL:
		b1 =  gen_linktype(ETHERTYPE_MOPDL);
		break;

	case Q_MOPRC:
		b1 =  gen_linktype(ETHERTYPE_MOPRC);
		break;

	default:
		abort();
	}
	return b1;
}

static struct block *
gen_ipfrag(void)
{
	struct slist *s;
	struct block *b;
	int reg1, reg2;

	reg1 = alloc_reg();
	reg2 = alloc_reg();
	/* not ip frag */
	s = new_stmt(BPF_LD|BPF_H|BPF_ABS, 0, 0, reg1, linkinfo.hdrlen + 6);
	sappend(s, new_stmt(BPF_ALU|BPF_AND|BPF_IMM, reg1, 0, reg2, 0x1fff));
	b = new_block(JMP(BPF_JEQ), reg2, 0);
	b->stmts = s;
#ifdef notdef
	gen_not(b);
#endif

	return b;
}

/*
 * Generate a statement list that parses the IP header
 * and places the offset of the encapsulating protocol in r1.
 * r0 is clobbered.
 */
static struct slist*
gen_ip_parse(int regno)
{
	struct slist *s;
	int reg1, reg2;

	reg1 = alloc_reg();
	reg2 = alloc_reg();
	s = new_stmt(BPF_LD|BPF_ABS|BPF_B, 0, 0, reg1, linkinfo.hdrlen);
	sappend(s, new_stmt(BPF_ALU|BPF_AND|BPF_IMM, reg1, 0, reg2, 0xf));
	sappend(s, new_stmt(BPF_ALU|BPF_LSH|BPF_IMM, reg2, 0, regno, 2));

	return (s);
}

static struct block *
gen_portatom(int off, long v)
{
	struct slist *s;
	struct block *b;
	int index = alloc_reg();
	int result = alloc_reg();

	/*
	 * Generate code to step over the IP header and inspect
	 * the port.
	 */
	s = gen_ip_parse(index);
	sappend(s, new_stmt(BPF_LD|BPF_IND|BPF_H, index, 0, result, linkinfo.hdrlen + off));
	b = new_block(JMP(BPF_JEQ), result, v);
	b->stmts = s;

	return b;
}

static struct block *
gen_portop(int port, int proto, int dir)
{
	struct block *b0, *b1, *tmp;

	/* ip proto 'proto' */
	tmp = gen_cmp(linkinfo.hdrlen + 9, BPF_B, (long)proto);
	b0 = gen_ipfrag();
	gen_and(tmp, b0);

	switch (dir) {
	case Q_SRC:
		b1 = gen_portatom(0, (long)port);
		break;

	case Q_DST:
		b1 = gen_portatom(2, (long)port);
		break;

	case Q_OR:
	case Q_DEFAULT:
		tmp = gen_portatom(0, (long)port);
		b1 = gen_portatom(2, (long)port);
		gen_or(tmp, b1);
		break;

	case Q_AND:
		tmp = gen_portatom(0, (long)port);
		b1 = gen_portatom(2, (long)port);
		gen_and(tmp, b1);
		break;

	default:
		abort();
	}
	gen_and(b0, b1);

	return b1;
}

static struct block *
gen_port(int port, int ip_proto, int dir)
{
	struct block *b0, *b1, *tmp;

	/* ether proto ip */
	b0 =  gen_linktype(ETHERTYPE_IP);

	switch (ip_proto) {
	case IPPROTO_UDP:
	case IPPROTO_TCP:
		b1 = gen_portop(port, ip_proto, dir);
		break;

	default:
		tmp = gen_portop(port, IPPROTO_TCP, dir);
		b1 = gen_portop(port, IPPROTO_UDP, dir);
		gen_or(tmp, b1);
		break;
	}
	gen_and(b0, b1);
	return b1;
}

static int
lookup_proto(char *name, int proto)
{
	int v;

	switch (proto) {
	case Q_DEFAULT:
	case Q_IP:
		v = pcap_nametoproto(name);
		if (v < 0)
			bpf_error("unknown ip proto '%s'", name);
		break;

	case Q_LINK:
		/* XXX should look up h/w protocol type based on linktype */
		v = pcap_nametoeproto(name);
		if (v < 0)
			bpf_error("unknown ether proto '%s'", name);
		break;

	default:
		v = -1;
		break;
	}
	return v;
}

static struct block *
gen_proto(int v, int proto, int dir)
{
	struct block *b0, *b1;

	if (dir != Q_DEFAULT)
		bpf_error("direction applied to 'proto'");

	switch (proto) {
	case Q_DEFAULT:
	case Q_IP:
		b0 = gen_linktype(ETHERTYPE_IP);
		b1 = gen_cmp(linkinfo.hdrlen + 9, BPF_B, (long)v);
		gen_and(b0, b1);
		return b1;

	case Q_ARP:
		bpf_error("arp does not encapsulate another protocol");
		/* NOTREACHED */

	case Q_RARP:
		bpf_error("rarp does not encapsulate another protocol");
		/* NOTREACHED */

	case Q_ATALK:
		bpf_error("atalk encapsulation is not specifiable");
		/* NOTREACHED */

	case Q_DECNET:
		bpf_error("decnet encapsulation is not specifiable");
		/* NOTREACHED */

	case Q_SCA:
		bpf_error("sca does not encapsulate another protocol");
		/* NOTREACHED */

	case Q_LAT:
		bpf_error("lat does not encapsulate another protocol");
		/* NOTREACHED */

	case Q_MOPRC:
		bpf_error("moprc does not encapsulate another protocol");
		/* NOTREACHED */

	case Q_MOPDL:
		bpf_error("mopdl does not encapsulate another protocol");
		/* NOTREACHED */

	case Q_LINK:
		return gen_linktype(v);

	case Q_UDP:
		bpf_error("'udp proto' is bogus");
		/* NOTREACHED */

	case Q_TCP:
		bpf_error("'tcp proto' is bogus");
		/* NOTREACHED */

	case Q_ICMP:
		bpf_error("'icmp proto' is bogus");
		/* NOTREACHED */

	case Q_IGMP:
		bpf_error("'igmp proto' is bogus");
		/* NOTREACHED */

	case Q_IGRP:
		bpf_error("'igrp proto' is bogus");
		/* NOTREACHED */

	default:
		abort();
		/* NOTREACHED */
	}
	/* NOTREACHED */
}

/*
 * Left justify 'addr' and return its resulting network mask.
 */
static u_long
net_mask(u_long *addr)
{
	register u_long m = 0xffffffff;

	if (*addr)
		while ((*addr & 0xff000000) == 0)
			*addr <<= 8, m <<= 8;

	return m;
}

struct block *
gen_scode(name, q)
	char *name;
	struct qual q;
{
	int proto = q.proto;
	int dir = q.dir;
	u_char *eaddr;
	u_long mask, addr, **alist;
	struct block *b, *tmp;
	int port, real_proto;

	switch (q.addr) {

	case Q_NET:
		addr = pcap_nametonetaddr(name);
		if (addr == 0)
			bpf_error("unknown network '%s'", name);
		mask = net_mask(&addr);
		return gen_host(addr, mask, proto, dir);

	case Q_DEFAULT:
	case Q_HOST:
		if (proto == Q_LINK) {
			eaddr = pcap_ether_hostton(name);
			if (eaddr == NULL)
				bpf_error(
				    "unknown ether host '%s'", name);
			return gen_ehostop(eaddr, dir);
		} else if (proto == Q_DECNET) {
			unsigned short dn_addr = __pcap_nametodnaddr(name);
			/*
			 * I don't think DECNET hosts can be multihomed, so
			 * there is no need to build up a list of addresses
			 */
			return (gen_host(dn_addr, 0, proto, dir));
		} else {
			alist = pcap_nametoaddr(name);
			if (alist == NULL || *alist == NULL)
				bpf_error("unknown host '%s'", name);
			b = gen_host(**alist++, 0xffffffffL, proto, dir);
			while (*alist) {
				tmp = gen_host(**alist++, 0xffffffffL,
					       proto, dir);
				gen_or(b, tmp);
				b = tmp;
			}
			return b;
		}

	case Q_PORT:
		if (proto != Q_DEFAULT && proto != Q_UDP && proto != Q_TCP)
			bpf_error("illegal qualifier of 'port'");
		if (pcap_nametoport(name, &port, &real_proto) == 0)
			bpf_error("unknown port '%s'", name);
		if (proto == Q_UDP) {
			if (real_proto == IPPROTO_TCP)
				bpf_error("port '%s' is tcp", name);
			else
				/* override PROTO_UNDEF */
				real_proto = IPPROTO_UDP;
		}
		if (proto == Q_TCP) {
			if (real_proto == IPPROTO_UDP)
				bpf_error("port '%s' is udp", name);
			else
				/* override PROTO_UNDEF */
				real_proto = IPPROTO_TCP;
		}
		return gen_port(port, real_proto, dir);

	case Q_GATEWAY:
		eaddr = pcap_ether_hostton(name);
		if (eaddr == NULL)
			bpf_error("unknown ether host: %s", name);

		alist = pcap_nametoaddr(name);
		if (alist == NULL || *alist == NULL)
			bpf_error("unknown host '%s'", name);
		return gen_gateway(eaddr, alist, proto, dir);

	case Q_PROTO:
		real_proto = lookup_proto(name, proto);
		if (real_proto >= 0)
			return gen_proto(real_proto, proto, dir);
		else
			bpf_error("unknown protocol: %s", name);

	case Q_UNDEF:
		syntax();
		/* NOTREACHED */
	}
	abort();
	/* NOTREACHED */
}

struct block *
gen_mcode(s1, s2, masklen, q)
	register const char *s1, *s2;
	register int masklen;
	struct qual q;
{
	register int nlen, mlen;
	u_long n, m;

	nlen = __pcap_atoin(s1, &n);
	/* Promote short ipaddr */
	n <<= 32 - nlen;

	if (s2 != NULL) {
		mlen = __pcap_atoin(s2, &m);
		/* Promote short ipaddr */
		m <<= 32 - mlen;
	} else {
		/* Convert mask len to mask */
		if (masklen > 32)
			bpf_error("mask length must be <= 32");
		m = 0xffffffff << (32 - masklen);
	}

	switch (q.addr) {

	case Q_NET:
		return gen_host(n, m, q.proto, q.dir);

	default:
		bpf_error("Mask syntax for networks only");
		/* NOTREACHED */
	}
}

struct block *
gen_ncode(s, v, q)
	register const char *s;
	u_long v;
	struct qual q;
{
	u_long mask;
	int proto = q.proto;
	int dir = q.dir;
	register int vlen;

	if (s == NULL)
		vlen = 32;
	else if (q.proto == Q_DECNET)
		vlen = __pcap_atodn(s, &v);
	else
		vlen = __pcap_atoin(s, &v);

	switch (q.addr) {

	case Q_DEFAULT:
	case Q_HOST:
	case Q_NET:
		if (proto == Q_DECNET)
			return gen_host(v, 0, proto, dir);
		else if (proto == Q_LINK) {
			bpf_error("illegal link layer address");
		} else {
			mask = net_mask(&v);
			return gen_host(v, mask, proto, dir);
		}

	case Q_PORT:
		if (proto == Q_UDP)
			proto = IPPROTO_UDP;
		else if (proto == Q_TCP)
			proto = IPPROTO_TCP;
		else if (proto == Q_DEFAULT)
			proto = -1;
		else
			bpf_error("illegal qualifier of 'port'");

		return gen_port((int)v, proto, dir);

	case Q_GATEWAY:
		bpf_error("'gateway' requires a name");
		/* NOTREACHED */

	case Q_PROTO:
		return gen_proto((int)v, proto, dir);

	case Q_UNDEF:
		syntax();
		/* NOTREACHED */

	default:
		abort();
		/* NOTREACHED */
	}
	/* NOTREACHED */
}

struct block *
gen_ecode(eaddr, q)
	u_char *eaddr;
	struct qual q;
{
	if ((q.addr == Q_HOST || q.addr == Q_DEFAULT) && q.proto == Q_LINK) {
		return gen_ehostop(eaddr, (int)q.dir);
	}
	bpf_error("ethernet address used in non-ether expression");
	/* NOTREACHED */
}

void
sappend(s0, s1)
	struct slist *s0, *s1;
{
	/*
	 * This is definitely not the best way to do this, but the
	 * lists will rarely get long.
	 */
	while (s0->next)
		s0 = s0->next;
	s0->next = s1;
}

struct arth *
gen_load(int proto, struct arth *index, int size)
{
	struct slist *s, *tmp;
	struct block *b;
	int regno, reg1, reg2;

	regno = alloc_reg();
	//free_reg(index->regno);
	switch (size) {

	default:
		bpf_error("data size must be 1, 2, or 4");

	case 1:
		size = BPF_B;
		break;

	case 2:
		size = BPF_H;
		break;

	case 4:
		size = BPF_W;
		break;
	}
	switch (proto) {
	default:
		bpf_error("unsupported index operation");

	case Q_LINK:
		s = new_stmt(BPF_LD|BPF_IND|size, index->regno, 0, regno, 0);
		sappend(index->s, s);
		break;

	case Q_IP:
	case Q_ARP:
	case Q_RARP:
	case Q_DECNET:
	case Q_LAT:
	case Q_MOPRC:
	case Q_MOPDL:
		/* XXX Note that we assume a fixed link link header here. */
		s = new_stmt(BPF_LD|BPF_IND|size, index->regno, 0, regno, linkinfo.hdrlen);
		sappend(index->s, s);

		b = gen_proto_abbrev(proto);
		if (index->b)
			gen_and(index->b, b);
		index->b = b;
		break;

	case Q_TCP:
	case Q_UDP:
	case Q_ICMP: {
	  int reg1, reg2, reg3, reg4;
	  
	  reg1 = alloc_reg();
	  reg2 = alloc_reg();
	  reg3 = alloc_reg();
	  reg4 = alloc_reg();

	  s = new_stmt(BPF_LD|BPF_ABS|BPF_B, 0, 0, reg1, linkinfo.hdrlen);
	  sappend(s, new_stmt(BPF_ALU|BPF_AND|BPF_IMM, reg1, 0, reg2, 0xf));
	  sappend(s, new_stmt(BPF_ALU|BPF_LSH|BPF_IMM, reg2, 0, reg3, 2));
	  sappend(s, new_stmt(BPF_ALU|BPF_ADD|BPF_REG,
			      index->regno, reg3, reg4, 0));
	  sappend(s, new_stmt(BPF_LD|BPF_IND|size,
			      reg4, reg4, regno, linkinfo.hdrlen));
	  sappend(index->s, s);
	  
	  gen_and(gen_proto_abbrev(proto), b = gen_ipfrag());
	  if (index->b)
	    gen_and(index->b, b);
	  index->b = b;
	  break;
	}
	}
	index->regno = regno;
	
	return index;
	
}

struct block *
gen_every(struct arth *num) {
  struct slist *s;
  struct block *b;
  int reg1, reg2, reg3, reg4, reg5;
  int memlocation;

  reg1 = alloc_reg();
  reg2 = alloc_reg();
  reg3 = alloc_reg();
  reg4 = alloc_reg();
  reg5 = alloc_reg();
  memlocation = alloc_mem();

  s = new_stmt(BPF_LDM, 0, 0, reg1, memlocation);
  sappend(s, new_stmt(BPF_ALU|BPF_ADD|BPF_IMM, reg1, 0, reg2, 1));
  sappend(s, new_stmt(BPF_ALU|BPF_DIV|BPF_REG, reg2, num->regno, reg3, 0));
  sappend(s, new_stmt(BPF_ALU|BPF_MUL|BPF_REG, reg3, num->regno, reg4, 0));
  sappend(s, new_stmt(BPF_ALU|BPF_SUB|BPF_REG, reg2, reg4, reg5, 0));
  sappend(s, new_stmt(BPF_STM, 0, 0, reg5, memlocation));
  sappend(num->s, s);

  b = new_block(JMP(BPF_JEQ), reg5, 0);
  b->stmts = num->s;

  return b;
}


struct block *
gen_relation(code, a0, a1, reversed)
	int code;
	struct arth *a0, *a1;
	int reversed;
{
	struct slist *s;
	struct block *b, *tmp;
	int regno;

	regno = alloc_reg();
	s = new_stmt(BPF_ALU|BPF_SUB|BPF_REG, a0->regno, a1->regno, regno, 0);
	b = new_block(JMP(code), regno, 0);
	if (code == BPF_JGT || code == BPF_JGE) {
		reversed = !reversed;
		b->s.imm = 0x80000000;
	}
	if (reversed)
		gen_not(b);

	sappend(a1->s, s);
	sappend(a0->s, a1->s);

	b->stmts = a0->s;

	//free_reg(a0->regno);
	//free_reg(a1->regno);

	/* 'and' together protocol checks */
	if (a0->b) {
		if (a1->b) {
			gen_and(a0->b, tmp = a1->b);
		}
		else
			tmp = a0->b;
	} else
		tmp = a1->b;

	if (tmp)
		gen_and(tmp, b);

	/*XXX don't bother freeing regno... keep it alive for CSE */

	return b;
}

/*XXX this doesn't work */
struct arth *
gen_loadlen()
{
	int regno = alloc_reg();
	int mem = alloc_mem();
	struct arth *a = (struct arth *)newchunk(sizeof(*a));
	struct slist *s;

	/*XXX r30 has length? */
	s = new_stmt(BPF_ALU|BPF_ADD|BPF_IMM, 30, 0, regno, 0);
	//s->next = new_stmt(BPF_STM, 0, 0, regno, mem);
	a->s = s;
	a->regno = regno;

	return a;
}

struct arth *
gen_loadi(val)
	int val;
{
	struct arth *a;
	struct slist *s;
	int reg;

	a = (struct arth *)newchunk(sizeof(*a));

	reg = alloc_reg();

	s = new_stmt(BPF_LD|BPF_IMM, 0, 0, reg, val);
#ifdef notdef
	//s->next = new_stmt(BPF_STM, 0, 0, 0, reg);
#endif
	a->s = s;
	a->regno = reg;

	return a;
}

struct arth *
gen_neg(a)
	struct arth *a;
{
	struct slist *s;

	s = new_stmt(BPF_ALU|BPF_NEG|BPF_IMM, a->regno, 0, a->regno, 0);
	sappend(a->s, s);

	return a;
}

struct arth *
gen_arth(code, a0, a1)
	int code;
	struct arth *a0, *a1;
{
	int regno;
	struct slist *s;

	regno = alloc_reg();
	s = new_stmt(BPF_ALU|BPF_REG|code, a0->regno, a1->regno, regno, 0);

	sappend(a1->s, s);
	sappend(a0->s, a1->s);

	/*XXX*/
	//free_reg(a1->regno);

	a0->regno = regno;

	return a0;
}

static int mem_max = 0;

/*
 * Memory used for persistent state across packet filters.
 * There is no free_mem(). 
 * 
 * Memory for register spilling is allocated after the end of this memory.
 */
static int alloc_mem() {
  return mem_max++;
}
  
int get_num_mem() {
  return mem_max + 1;
}


/*
 * Here we handle simple allocation of the scratch registers.
 * If too many registers are alloc'd, the allocator punts.
 */

static int regstack[MAX_PSEUDO_REGS];
static int reg_tos = 0;

static int reg_max = 0;

/*
 * Return the next free register.
 */
static int
alloc_reg()
{
	int reg;
	if (reg_tos > 0)
		return (regstack[--reg_tos]);
	if (reg_max >= MAX_PSEUDO_REGS)
		bpf_error("ran out of pseudo-regs");

	reg = reg_max;
	/* register 30 and 31 are reserved */
	if (reg == 30) { reg_max+=3; return reg+2; }
	if (reg == 31) { reg_max+=2; return reg+1; }
	reg_max++;
	return reg;
}

int get_num_reg() {
  return (reg_max > 31) ? reg_max + 1 : 32;
}


/*
 * Return a register to the table so it can
 * be used later.
 */
static void
free_reg(n)
	int n;
{
  regstack[reg_tos++] = n;
}

static struct block *
gen_len(int jmp, int n)
{
	struct slist *s;
	struct block *b;
	int reg = alloc_reg();

	s = new_stmt(BPF_ALU|BPF_ADD|BPF_IMM, 30, 0, reg, 0);
	b = new_block(JMP(jmp), reg, n);
	b->stmts = s;

	return b;
}

struct block *
gen_greater(n)
	int n;
{
	return gen_len(BPF_JGE, n);
}

struct block *
gen_less(n)
	int n;
{
	struct block *b;

	b = gen_len(BPF_JGT, n);
	gen_not(b);

	return b;
}

struct block *
gen_byteop(op, idx, val)
	int op, idx, val;
{
	struct block *b;
	struct slist *s;
	int reg;

	switch (op) {
	default:
		abort();

	case '=':
		return gen_cmp((u_int)idx, BPF_B, (long)val);

	case '<':
		b = gen_cmp((u_int)idx, BPF_B, (long)val);
		b->s.code = JMP(BPF_JGE);
		gen_not(b);
		return b;

	case '>':
		b = gen_cmp((u_int)idx, BPF_B, (long)val);
		b->s.code = JMP(BPF_JGT);
		return b;

	case '|':
	  reg = alloc_reg();
	  s = new_stmt(BPF_ALU|BPF_OR|BPF_IMM, 0, 0, reg, val);
	  break;

	case '&':
	  reg = alloc_reg();
	  s = new_stmt(BPF_ALU|BPF_AND|BPF_IMM, 0, 0, reg, val);
	  break;
	}
	b = new_block(JMP(BPF_JEQ), reg, 0);
	b->stmts = s;
	gen_not(b);

	return b;
}

struct block *
gen_broadcast(proto)
	int proto;
{
	u_long hostmask;
	struct block *b0, *b1, *b2;
	static u_char ebroadcast[] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

	switch (proto) {

	case Q_DEFAULT:
	case Q_LINK:
		return gen_ehostop(ebroadcast, Q_DST);
		break;

	case Q_IP:
		b0 = gen_linktype(ETHERTYPE_IP);
		hostmask = ~netmask;
		b1 = gen_mcmp(linkinfo.hdrlen + 16, BPF_W, (long)0, hostmask);
		b2 = gen_mcmp(linkinfo.hdrlen + 16, BPF_W,
			      (long)(~0 & hostmask), hostmask);
		gen_or(b1, b2);
		gen_and(b0, b2);
		return b2;
	}
	bpf_error("only ether/ip broadcast filters supported");
}

struct block *
gen_multicast(proto)
	int proto;
{
	register struct block *b0, *b1;
	register struct slist *s;
	int reg1, reg2, reg3;
	
	switch (proto) {

	case Q_DEFAULT:
	case Q_LINK:
		/* ether[0] & 1 != 0 */
	  reg1 = alloc_reg();
	  reg2 = alloc_reg();
	  
	  s = new_stmt(BPF_LD|BPF_B|BPF_ABS, 0, 0, reg1, 0);
	  sappend(s, new_stmt(BPF_ALU|BPF_AND|BPF_IMM,
			      reg1, 0, reg2, 1));
	  b0 = new_block(JMP(BPF_JEQ), reg2, 0);
	  b0->stmts = s;
	  gen_not(b0);
	  return b0;
	  
	case Q_IP:
		b0 = gen_linktype(ETHERTYPE_IP);
		b1 = gen_cmp(linkinfo.hdrlen + 16, BPF_B, (long)224);
		b1->s.code = JMP(BPF_JGE);
		gen_and(b0, b1);
		return b1;
	}
	bpf_error("link-level multicast filters only on ethernet/FDDI");
}

/*
 * generate command for inbound/outbound.  It's here so we can
 * make it link-type specific.  'dir' = 0 implies "inbound",
 * = 1 implies "outbound".
 */
struct block *
gen_inbound(dir)
	int dir;
{
	register struct block *b0;

	b0 = gen_relation(BPF_JEQ,
			  gen_load(Q_LINK, gen_loadi(0), 1),
			  gen_loadi(0),
			  dir);
	return (b0);
}
